from decimal import Decimal
from enum import Enum, auto
from typing import Any, Union

from loguru import logger
from tompy.decimal import di_down

from otlpy.base.net import AsyncHttpClient
from otlpy.base.order import (
    ORDER_TYPE,
    Buy,
    Cancel,
    Order,
    OrderAPI,
    Replace,
    Sell,
)
from otlpy.kis.common import Common


class US_OVRS_EXCG_CD(Enum):
    NASD = auto()
    NYSE = auto()
    AMEX = auto()


def us_str_order_type(order_type: ORDER_TYPE, /) -> str:
    if order_type == ORDER_TYPE.LIMIT:
        s = "00"
    else:
        assert False
    return s


class OverseasStock(OrderAPI):
    def __init__(self, common: Common, client: AsyncHttpClient, /) -> None:
        super().__init__(has_replace=True)
        self.__common = common
        self.__client = client

    @property
    def _common(self, /) -> Common:
        return self.__common

    @property
    def _client(self, /) -> AsyncHttpClient:
        return self.__client

    async def token(self, /) -> None:
        await self._common.token(self._client)
        logger.info("token")

    async def _new_order(
        self,
        order: Union[Buy, Sell],
        /,
        *,
        tr_id: str,
        excg_cd: str,
        ord_dvsn: str,
    ) -> None:
        common = self._common
        client = self._client
        url_path = "/uapi/overseas-stock/v1/trading/order"
        data = {
            "CANO": common.account_cano_overseas_stock,
            "ACNT_PRDT_CD": common.account_prdt_overseas_stock,
            "OVRS_EXCG_CD": excg_cd,
            "PDNO": order.ticker,
            "ORD_QTY": str(di_down(order.qty)),
            "OVRS_ORD_UNPR": str(order.price),
        }
        if ord_dvsn:
            data["ORD_DVSN"] = ord_dvsn
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "hashkey": await common.hash(client, data),
        }
        err, rheaders, rdata = await client.post(url_path, headers, data)
        if err:
            return
        if rdata["rt_cd"] != "0":
            logger.error(f"{url_path}\n{headers}\n{data}\n{rheaders}\n{rdata}")
            return
        order.acknowledgment(rdata, rdata["output"]["ODNO"], order.qty)

    async def _cancel_or_replace_order(
        self,
        order: Union[Cancel, Replace],
        /,
        *,
        tr_id: str,
        excg_cd: str,
        omsg: str,
        ack_qty: Decimal,
    ) -> None:
        common = self._common
        client = self._client
        rdata1 = order.origin.rdata["output"]
        url_path = "/uapi/overseas-stock/v1/trading/order-rvsecncl"
        data = {
            "CANO": common.account_cano_overseas_stock,
            "ACNT_PRDT_CD": common.account_prdt_overseas_stock,
            "OVRS_EXCG_CD": excg_cd,
            "PDNO": order.ticker,
            "ORGN_ODNO": rdata1["ODNO"],
            "RVSE_CNCL_DVSN_CD": omsg,
            "ORD_QTY": str(di_down(order.qty)),
            "OVRS_ORD_UNPR": str(order.price),
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "hashkey": await common.hash(client, data),
        }
        err, rheaders, rdata = await client.post(url_path, headers, data)
        if err:
            return
        if rdata["rt_cd"] != "0":
            logger.error(f"{url_path}\n{headers}\n{data}\n{rheaders}\n{rdata}")
            return
        order.acknowledgment(rdata, rdata["output"]["ODNO"], ack_qty)

    async def _all_orders(
        self,
        yyyymmdd: str,
        /,
        *,
        tr_cont: str,
        ctx_area_fk200: str,
        ctx_area_nk200: str,
        tr_id: str,
        excg_cd: str,
        ticker: str = "",
    ) -> tuple[bool, Any]:
        common = self._common
        url_path = "/uapi/overseas-stock/v1/trading/inquire-ccnl"
        data = {
            "CANO": common.account_cano_overseas_stock,
            "ACNT_PRDT_CD": common.account_prdt_overseas_stock,
            "PDNO": ticker,
            "ORD_STRT_DT": yyyymmdd,
            "ORD_END_DT": yyyymmdd,
            "SLL_BUY_DVSN": "00",
            "CCLD_NCCS_DVSN": "00",
            "OVRS_EXCG_CD": excg_cd,
            "SORT_SQN": "AS",
            "ORD_DT": "",
            "ORD_GNO_BRNO": "",
            "ODNO": "",
            "CTX_AREA_NK200": ctx_area_nk200,
            "CTX_AREA_FK200": ctx_area_fk200,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
            "tr_cont": tr_cont,
        }
        err, rheaders, rdata = await self._client.get(url_path, headers, data)
        if err:
            return False, None
        if rdata["rt_cd"] != "0":
            logger.error(f"{url_path}\n{headers}\n{data}\n{rheaders}\n{rdata}")
            return False, None
        if rheaders["tr_cont"] == "F" or rheaders["tr_cont"] == "M":
            return True, rdata
        if rheaders["tr_cont"] == "D" or rheaders["tr_cont"] == "E":
            return False, rdata
        assert False

    async def _price(self, ticker: str, /, *, excd: str) -> Any:
        common = self._common
        tr_id = "HHDFS00000300"
        url_path = "/uapi/overseas-price/v1/quotations/price"
        data = {
            "AUTH": "",
            "EXCD": excd,
            "SYMB": ticker,
        }
        headers = {
            **common.headers4(),
            "tr_id": tr_id,
        }
        err, rheaders, rdata = await self._client.get(url_path, headers, data)
        if err:
            return None
        if rdata["rt_cd"] != "0":
            logger.error(f"{url_path}\n{headers}\n{data}\n{rheaders}\n{rdata}")
            return None
        return rdata


class NASD(OverseasStock):
    async def buy(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Buy:
        order = Buy(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1002U",
            excg_cd="NASD",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def sell(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Sell:
        order = Sell(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1006U",
            excg_cd="NASD",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def cancel(self, origin: Order, /) -> Cancel:
        order = Cancel(origin, ORDER_TYPE.LIMIT)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="NASD",
            omsg="02",
            ack_qty=Decimal(),
        )
        return order

    async def replace(
        self, origin: Order, order_type: ORDER_TYPE, /, price: Decimal
    ) -> Replace:
        order = Replace(origin, order_type, price)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="NASD",
            omsg="01",
            ack_qty=order.qty,
        )
        return order

    async def all_orders(
        self,
        yyyymmdd: str,
        /,
        *,
        tr_cont: str,
        ctx_area_fk200: str,
        ctx_area_nk200: str,
        ticker: str = "",
    ) -> tuple[bool, Any]:
        return await super()._all_orders(
            yyyymmdd,
            tr_cont=tr_cont,
            ctx_area_fk200=ctx_area_fk200,
            ctx_area_nk200=ctx_area_nk200,
            tr_id="JTTT3001R",
            excg_cd="NASD",
            ticker=ticker,
        )

    async def price(self, ticker: str, /) -> Any:
        return await super()._price(ticker, excd="NAS")


class NYSE(OverseasStock):
    async def buy(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Buy:
        order = Buy(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1002U",
            excg_cd="NYSE",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def sell(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Sell:
        order = Sell(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1006U",
            excg_cd="NYSE",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def cancel(self, origin: Order, /) -> Cancel:
        order = Cancel(origin, ORDER_TYPE.LIMIT)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="NYSE",
            omsg="02",
            ack_qty=Decimal(),
        )
        return order

    async def replace(
        self, origin: Order, order_type: ORDER_TYPE, /, price: Decimal
    ) -> Replace:
        order = Replace(origin, order_type, price)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="NYSE",
            omsg="01",
            ack_qty=order.qty,
        )
        return order

    async def all_orders(
        self,
        yyyymmdd: str,
        /,
        *,
        tr_cont: str,
        ctx_area_fk200: str,
        ctx_area_nk200: str,
        ticker: str = "",
    ) -> tuple[bool, Any]:
        return await super()._all_orders(
            yyyymmdd,
            tr_cont=tr_cont,
            ctx_area_fk200=ctx_area_fk200,
            ctx_area_nk200=ctx_area_nk200,
            tr_id="JTTT3001R",
            excg_cd="NYSE",
            ticker=ticker,
        )

    async def price(self, ticker: str, /) -> Any:
        return await super()._price(ticker, excd="NYS")


class AMEX(OverseasStock):
    async def buy(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Buy:
        order = Buy(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1002U",
            excg_cd="AMEX",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def sell(
        self,
        order_type: ORDER_TYPE,
        ticker: str,
        /,
        qty: Decimal,
        price: Decimal,
    ) -> Sell:
        order = Sell(order_type, ticker, qty, price)
        await self._new_order(
            order,
            tr_id="JTTT1006U",
            excg_cd="AMEX",
            ord_dvsn=us_str_order_type(order.otype),
        )
        return order

    async def cancel(self, origin: Order, /) -> Cancel:
        order = Cancel(origin, ORDER_TYPE.LIMIT)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="AMEX",
            omsg="02",
            ack_qty=Decimal(),
        )
        return order

    async def replace(
        self, origin: Order, order_type: ORDER_TYPE, /, price: Decimal
    ) -> Replace:
        order = Replace(origin, order_type, price)
        await self._cancel_or_replace_order(
            order,
            tr_id="JTTT1004U",
            excg_cd="AMEX",
            omsg="01",
            ack_qty=order.qty,
        )
        return order

    async def all_orders(
        self,
        yyyymmdd: str,
        /,
        *,
        tr_cont: str,
        ctx_area_fk200: str,
        ctx_area_nk200: str,
        ticker: str = "",
    ) -> tuple[bool, Any]:
        return await super()._all_orders(
            yyyymmdd,
            tr_cont=tr_cont,
            ctx_area_fk200=ctx_area_fk200,
            ctx_area_nk200=ctx_area_nk200,
            tr_id="JTTT3001R",
            excg_cd="AMEX",
            ticker=ticker,
        )

    async def price(self, ticker: str, /) -> Any:
        return await super()._price(ticker, excd="AMS")
